#include "includes.h"
#include "knightcap.h"

/*
Here are some samples of what FICS looks like:

fics% match 
jeromed (1601) seeking 2 12 unrated blitz ("play 17" to respond)
fics% ThingOne
Issuing: tridge (++++) ThingOne (2247) unrated blitz 2 12.
--** ThingOne is a computer **--
fics% 
ThingOne accepts your challenge.
fics% 
Creating: ThingOne (2247) tridge (++++) unrated blitz 2 12
{Game 34 (ThingOne vs. tridge) Creating unrated blitz match.}

#@#000tango           *tridge          :RNBQKBNRPPPP PPP            P                   pppppppprnbqkbnr001B39390011500120P/e2-e4(0:17)@#@



fics% 
Challenge: ixosjaps (1935) Tridge (2128) rated blitz 2 12.
Your blitz rating will change:  Win: +33,  Draw: -32,  Loss: -98
Your new RD will be 151.7
You can "accept" or "decline", or propose different parameters.
fics% acceptposition is size 224

You accept the challenge of ixosjaps.

Creating: Tridge (2128) ixosjaps (1935) rated blitz 2 12
{Game 14 (Tridge vs. ixosjaps) Creating rated blitz match.}

Game 14 (Tridge vs. ixosjaps)

#@#013Tridge          *ixosjaps        :RNBQKBNRPPPPPPPP                                pppppppprnbqkbnr001W39390012000120none   (0:00)@#@
fics% W: 20 n/s 105/0 2:2 mo=21/2 hash=0/105 t=0 d2d3
W: -12 n/s 639/0 3:4 mo=80/21 hash=65/641 t=0 d2d3


blindfolded accepts your challenge.
fics% 
Creating: blindfolded (1687) KnightCap (1992) rated blitz 2 10
{Game 6 (blindfolded vs. KnightCap) Creating rated blitz match.}

Game 6 (blinposition is size 228


White (KnightCap) : 2 mins, 16 secs
Black (mrb) : 1 mins, 34 secs

Notification: TayxBot, who has an adjourned game with you, has arrived.
fics% 
Challenge: TayxBot (1940) KnightCap (1940) rated blitz 2 10 (adjourned).
--** TayxBot is a computer **--
Your blitz rating will change:  Win: +4,  Draw: +0,  Loss: -4

{Game 52 (KlausK vs. KnightCap) KlausK resigns} 0-1
*/

extern char *ics_name;

#define ICSNAME ics_name
#define ICSPROMPT "fics% "

void ics_thanks(void)
{
	prog_printf("say thanks for the game\n");
}

static void parse_ics_special(char *line)
{
	char white_player[50];
	char black_player[50];
	char *opponent;
	char white_rating[50], black_rating[50];
	int game_num;
	char dum[100], result[100];
	int increment, game_time;
	static FILE *f;
	char *tag;

	extern struct state *state;

	if (!state->ics_robot)
		return;

	if (!f) {
		f = fopen("ics.log", "w");
		if (!f)
			f = fopen("/tmp/ics.log", "w");
		setlinebuf(f);
	}

	/* maybe we think we are playing, but we're not */
	tag = "You are neither playing";
	if (strncmp(line,tag,strlen(tag)) == 0) {
		state->computer = 0;
		return;
	}

	/* is it all over? */
	if (sscanf(line,"{Game %d %s vs. %s %s %s", 
		   &game_num, white_player, black_player, dum, result) == 5 &&
	    strstr(line,ics_name)) {
		if (strstr(line,"1-0") || strstr(line,"1/2-1/2") ||
		    strstr(line,"0-1")) {
			ics_thanks();
			state->computer = 0;
		}
		return;
	}

	tag = "Tridge tells you: icsrun ";
	if (strncmp(line,tag, strlen(tag)) == 0) {
		if (!parse_stdin(line + strlen(tag))) {
			if (prog_running()) {
				prog_printf("%s\n", line + strlen(tag));
			}
		}
		return;
	}

	/* has someone challenged us? */
	if (sscanf(line,"Challenge: %s", dum) == 1) {
		/* auto accept! */
		prog_printf("accept\n");
		fprintf(f, "accepted challenge\n");
		return;
	}

	/* has a match been created? */
	if (sscanf(line,"Creating: %s %s %s %s %s %s %d %d",
		   white_player, white_rating,
		   black_player, black_rating,
		   dum, dum,
		   &game_time, &increment) == 8) {

		reset_board();
		if (strcasecmp(white_player, ICSNAME) == 0) {
			state->computer = PIECE_WHITE;
			state->flip_view = 0;
			opponent = black_player;
		} else {
			state->flip_view = 1;
			state->computer = PIECE_BLACK;
			opponent = white_player;
		}

		timer_estimate(game_time*60, increment);

		prog_printf("kibitz Hello from KnightCap! (running in ICS robot mode)\n");
		fprintf(f, "Creating %s vs %s mt=%d\n",
			white_player, black_player, state->move_time);
		return;
	}
}


int parse_ics_move(char *line,int player, Move *move, Piece *promotion)
{
	extern struct state *state;
	int nextplayer, x, y;
	char board[20][9]; /* its made too long to be less error prone */
	char tomove;
	int pawnfile, wshort, wlong, bshort, blong, fiftycount, game;
	char whiteplayer[100], blackplayer[100];
	int relation;
	int initial_time;
	int increment;
	int white_time, black_time;
	int movenum;
	char prevmove[20];
	Position b1, b2;
	Position *b = &state->position;
	int bad_move = 0;
	int wstrength, bstrength;
	int from, to;

	while (*line && !isprint(*line))
		line++;

	while (strncmp(line,ICSPROMPT,strlen(ICSPROMPT)) == 0)
		line += strlen(ICSPROMPT);

	memset(board, 0, sizeof(board));

	if (sscanf(line,"<12> %s %s %s %s %s %s %s %s %c %d %d %d %d %d %d %d %s %s %d %d %d %d %d %d %d %d %s",
		   board[7], board[6], board[5], board[4], 
		   board[3], board[2], board[1], board[0], 
		   &tomove,
		   &pawnfile,
		   &wshort, &wlong,
		   &bshort, &blong,
		   &fiftycount,
		   &game,
		   whiteplayer, blackplayer,
		   &relation,
		   &initial_time,
		   &increment,
		   &wstrength,
		   &bstrength,
		   &white_time,
		   &black_time,
		   &movenum,
		   prevmove) != 27) {
		parse_ics_special(line);
		return 0;
	}

	if (tomove == 'W') {
		nextplayer = 1;
	} else if (tomove == 'B') {
		nextplayer = -1;	
	} else {
		return 0;
	}

	if (strcmp(ics_name, whiteplayer) == 0) {
		timer_estimate(white_time, increment);
		if (black_time <= 0)
			prog_printf("flag\n");
	}

	if (strcmp(ics_name, blackplayer) == 0) {
		timer_estimate(black_time, increment);
		if (white_time <= 0)
			prog_printf("flag\n");
	}

	if (tomove == 'W') {
		lprintf(0,"*%s: %d:%02d    %s %d:%02d    %d. ... %s    mtime=%d\n", 
			whiteplayer, white_time/60, white_time%60,
			blackplayer, black_time/60, black_time%60,
			movenum, prevmove, state->move_time);
	} else {
		lprintf(0," %s: %d:%02d   *%s %d:%02d    %d. %s        mtime=%d\n", 
			whiteplayer, white_time/60, white_time%60,
			blackplayer, black_time/60, black_time%60,
			movenum, prevmove, state->move_time);
	}

	if (!strncmp(prevmove, "none", 4)) {
		/* its a new game */
		reset_board();
		if (abs(relation) == 1 && state->ics_robot) {
			if (strcmp(whiteplayer, ics_name) == 0) {
				state->computer = 1;
			} else {
				state->computer = -1;
			}
		}
		return 0;
	} else if (!strncmp(prevmove, "o-o-o", 5)) {
		if (nextplayer < 0) {
			move->from = E1;
			move->to = C1;
		} else {
			move->from = E8;
			move->to = C8;
		}
	} else if (!strncmp(prevmove, "o-o", 3)) {
		if (nextplayer < 0) {
			move->from = E1;
			move->to = G1;
		} else {
			move->from = E8;
			move->to = G8;
		}
	} else {
		if (prevmove[1] != '/') {
			return 0;
		}
		
		if (prevmove[4] != '-') {
			return 0;
		}

		if (!parse_square(prevmove+2, &from) || 
		    !parse_square(prevmove+5, &to))
			return 0;

		move->from = from;
		move->to = to;
	}

	if (!legal_move(&state->position, move) ||
	    nextplayer != -player) {
		bad_move = 1;
	} 

	memset(&b1, 0, sizeof(b1));

	for (x=0;x<8;x++)
		for (y=0;y<8;y++) {
			Piece p;
			p = charpiece(board[y][x]);
			if (islower(board[y][x])) p = -p;
			b1.board[POSN(x,y)] = p;
		}

	if (memcmp(b->board, b1.board, sizeof(b1.board)) == 0) {
		/* we already have the right position */
		return 0;
	}

	if (wshort)
		b1.flags |= WHITE_CASTLE_SHORT;
	if (bshort)
		b1.flags |= BLACK_CASTLE_SHORT;
	if (wlong)
		b1.flags |= WHITE_CASTLE_LONG;
	if (blong)
		b1.flags |= BLACK_CASTLE_LONG;
	
	if (pawnfile != -1) {
		if (nextplayer == -1)
			b1.enpassent = POSN(pawnfile, 2);
		else
			b1.enpassent = POSN(pawnfile, 5);
	}

	b1.fifty_count = fiftycount;
	
	if (!bad_move) {
		if (!do_move(&b2, b, move)) {
			bad_move = 1;
		}
		if (memcmp(b2.board, b1.board, sizeof(b1.board)) != 0) {
			bad_move = 1;
		}
	}

	if (bad_move) {
		/* assume ICS is right! We try to generate a valid
		   position from what its given us */
		state->computer = 0;
		memset(state->game_record, 0, sizeof(state->game_record));
		memset(state->hash_list, 0, sizeof(state->hash_list));

		(*b) = b1;

		b->num_moves = ((movenum-1) * 2);
		if (nextplayer == -1)
			b->num_moves++;

		b->hash1 = random();
		b->hash2 = random();

		/* now the tricky bit - getting the pboard and piece list
		   reasonable */
		create_pboard(b);

		if (abs(relation) == 1 && state->ics_robot) {
			if (strcmp(whiteplayer, ics_name) == 0) {
				state->computer = 1;
			} else {
				state->computer = -1;
			}
		}

		lprintf(0,"generated position\n");
		return 0;
	}

	/* its all ok! */

	return 1;
}


